@@ -3,9 +3,10 @@ a design matrix and converting to an OrderedDict that can be
 read by semeio.fmudesign.DesignMatrix.generate
 """
 
-from collections import OrderedDict
-from collections.abc import Mapping
-from typing import Any
+from collections import Counter, OrderedDict
+from collections.abc import Hashable, Mapping, Sequence
+from pathlib import Path
+from typing import Any, cast
 
 import numpy as np
 import openpyxl
@@ -14,7 +15,11 @@ import yaml
 
 
 def excel2dict_design(
-    input_filename: str, sheetnames: Mapping[str, Any] | None = None
+    input_filename: str,
+    *,
+    gen_input_sheet: str = "general_input",
+    design_input_sheet: str = "designinput",
+    default_val_sheet: str = "defaultvalues",
 ) -> OrderedDict[str, Any]:
     """Read excel file with input to design setup
     Currently only specification of
@@ -22,39 +27,46 @@ def excel2dict_design(
 
     Args:
         input_filename (str): Name of excel input file
-        sheetnames (dict): Dictionary of worksheet names to load
-            information from. Supported keys: general_input, defaultvalues,
-            and designinput.
+        gen_input_sheet (str): Sheet name for general input
+        design_input_sheet (str): Sheet name for design input
+        default_val_sheet (str): Sheet name for default input
 
     Returns:
         OrderedDict on format for DesignMatrix.generate
     """
-    if not sheetnames or "general_input" not in sheetnames:
-        gen_input_sheet = _find_geninput_sheetname(input_filename)
-    else:
-        gen_input_sheet = sheetnames["general_input"]
 
-    generalinput = pd.read_excel(
-        input_filename, gen_input_sheet, header=None, index_col=0, engine="openpyxl"
+    # Find sheets
+    xlsx = openpyxl.load_workbook(input_filename, read_only=True, keep_links=False)
+    gen_input_sheet = find_sheet(gen_input_sheet, names=xlsx.sheetnames)
+    design_input_sheet = find_sheet(design_input_sheet, names=xlsx.sheetnames)
+    default_val_sheet = find_sheet(default_val_sheet, names=xlsx.sheetnames)
+
+    generalinput = (
+        pd.read_excel(
+            input_filename, gen_input_sheet, header=None, index_col=0, engine="openpyxl"
+        )
+        .dropna(axis=0, how="all")
+        .dropna(axis=1, how="all")
     )
-    generalinput.dropna(axis=0, how="all", inplace=True)
-    generalinput.dropna(axis=1, how="all", inplace=True)
 
-    if str(generalinput[1]["designtype"]) == "onebyone":
-        returndict = _excel2dict_onebyone(input_filename, sheetnames)
-    elif "seeds" in generalinput[1]:
+    design_type = str(generalinput.loc["designtype"].iloc[0])
+    if design_type != "onebyone":
+        raise ValueError(
+            f"Generation of DesignMatrix only implemented for type 'onebyone', not {design_type}"
+        )
+
+    if "seeds" in generalinput.index:
         raise ValueError(
             "The 'seeds' parameter has been deprecated and is no longer supported. "
             "Use 'rms_seeds' instead"
         )
-    else:
-        raise ValueError(
-            "Generation of DesignMatrix only "
-            "implemented for type onebyone "
-            "In general_input designtype was "
-            "set to {}".format(str(generalinput[1]["designtype"]))
-        )
-    return returndict
+
+    return _excel2dict_onebyone(
+        input_filename=input_filename,
+        gen_input_sheet=gen_input_sheet,
+        design_input_sheet=design_input_sheet,
+        default_val_sheet=default_val_sheet,
+    )
 
 
 def inputdict_to_yaml(inputdict: Mapping[str, Any], filename: str) -> None:
@@ -68,126 +80,61 @@ def inputdict_to_yaml(inputdict: Mapping[str, Any], filename: str) -> None:
         yaml.dump(inputdict, stream)
 
 
-def _find_geninput_sheetname(input_filename: str) -> str:
-    """Finding general input sheet, allowing for name
-    variations."""
-    xlsx = openpyxl.load_workbook(input_filename, read_only=True, keep_links=False)
-    sheets = xlsx.sheetnames
-    general_input_sheet: list[str] = []
-    for sheet in sheets:
-        if sheet in [
-            "general_input",
-            "generalinput",
-            "GeneralInput",
-            "Generalinput",
-            "General_Input",
-            "General_input",
-        ]:
-            general_input_sheet.append(sheet)
+def find_sheet(name: str, names: list[str]) -> str:
+    """Search for Excel sheets with a soft matching. Raises ValueError if zero
+    or more than one match is found.
 
-    if len(general_input_sheet) > 1:
-        raise ValueError(
-            f"More than one sheet with general input. Sheetnames are {general_input_sheet} "
-        )
-
-    if not general_input_sheet:
-        raise ValueError(
-            f"No general_input sheet provided in Excel file {input_filename} "
-        )
-
-    return general_input_sheet[0]
-
-
-def _find_onebyone_defaults_sheet(input_filename: str) -> str:
-    """Finds correct sheet name for default values to use when parsing
-    excel file.
-
-    Returns:
-        string, name of a sheet in the excel file
+    Examples:
+    >>> find_sheet('general_input', ['generalinput', 'designinput', 'defaultinput'])
+    'generalinput'
+    >>> find_sheet('variable_input', ['generalinput', 'designinput', 'defaultinput'])
+    Traceback (most recent call last):
+      ...
+    ValueError: No match for variable_input: ['generalinput', 'designinput', 'defaultinput']
     """
-    xlsx = openpyxl.load_workbook(input_filename, read_only=True, keep_links=False)
-    sheets = xlsx.sheetnames
 
-    default_values_sheet: list[str] = []
+    def sanitize(inputstring: str) -> str:
+        return inputstring.lower().strip().replace("_", "")
 
-    for sheet in sheets:
-        if sheet in [
-            "default_values",
-            "defaultvalues",
-            "DefaultValues",
-            "Defaultvalues",
-            "Default_Values",
-            "Default_values",
-        ]:
-            default_values_sheet.append(sheet)
-    if len(default_values_sheet) > 1:
-        raise ValueError(
-            f"More than one sheet with default values. Sheetnames are {default_values_sheet} "
-        )
-
-    if not default_values_sheet:
-        raise ValueError(
-            f"No defaultvalues sheet provided in Excel file {input_filename} "
-        )
-
-    return default_values_sheet[0]
-
-
-def _find_onebyone_input_sheet(input_filename: str) -> str:
-    """Finds correct sheet name for input to use when parsing excel file.
-
-    Returns:
-        string, name of a sheet in the excel file
-    """
-    xlsx = openpyxl.load_workbook(input_filename, read_only=True, keep_links=False)
-    sheets = xlsx.sheetnames
-
-    design_input_sheet: list[str] = []
-
-    for sheet in sheets:
-        if sheet in [
-            "design_input",
-            "designinput",
-            "DesignInput",
-            "Designinput",
-            "Design_Input",
-            "Design_input",
-        ]:
-            design_input_sheet.append(sheet)
-    if len(design_input_sheet) > 1:
-        raise ValueError(
-            f"More than one sheet with design inputSheetnames are {design_input_sheet} "
-        )
-
-    if not design_input_sheet:
-        raise ValueError(
-            f"No designinput sheet provided in Excel file {input_filename} "
-        )
-    return design_input_sheet[0]
+    found = [name_i for name_i in names if sanitize(name) == sanitize(name_i)]
+    if len(found) > 1:
+        raise ValueError(f"More than one match for {name}: {found}")
+    elif len(found) == 0:
+        raise ValueError(f"No match for {name}: {names}")
+    else:
+        return found[0]
 
 
 def _check_designinput(dsgn_input: pd.DataFrame) -> None:
     """Checks for valid input in designinput sheet"""
+    # Filter out rows where sensname has no value
+    valid_sensnames = dsgn_input["sensname"].dropna()
+    duplicated_mask = valid_sensnames.duplicated()
 
-    # Check for duplicate sensnames
-    sensitivity_names = []
-    for row in dsgn_input.itertuples():
-        if _has_value(row.sensname):
-            if row.sensname in sensitivity_names:
-                raise ValueError(
-                    f"sensname '{row.sensname}' was found on more than one row in designinput "
-                    "sheet. Two sensitivities can not share the same sensname. "
-                    "Please correct this and rerun"
-                )
-            sensitivity_names.append(row.sensname)
+    if duplicated_mask.any():
+        # Find the first duplicate to include in error message
+        duplicate_name = valid_sensnames[duplicated_mask].iloc[0]
+        raise ValueError(
+            f"sensname '{duplicate_name}' was found on more than one row in designinput "
+            "sheet. Two sensitivities can not share the same sensname. "
+            "Please correct this and rerun"
+        )
+
+    # Check for duplicate parameter names within each sensname
+    for sensname, df_sensname in dsgn_input.ffill().groupby("sensname"):
+        param_names = list(df_sensname["param_name"])
+        try:
+            _raise_if_duplicates(param_names)
+        except ValueError as e:
+            raise ValueError(f"Duplicate param names in {sensname}\n{e}") from e
 
 
 def _check_for_mixed_sensitivities(sens_name: str, sens_group: pd.DataFrame) -> None:
     """Checks for valid input in designinput sheet. A sensitivity cannot contain
     two different sensitivity types"""
 
-    types = sens_group.groupby("type", sort=False)
-    if len(types) > 1:
+    unique_types = sens_group["type"].dropna().unique()
+    if len(unique_types) > 1:
         raise ValueError(
             f"The sensitivity with sensname '{sens_name}' in designinput sheet contains more "
             "than one sensitivity type. For each sensname all parameters must be "
@@ -196,13 +143,41 @@ def _check_for_mixed_sensitivities(sens_name: str, sens_group: pd.DataFrame) ->
         )
 
 
+def resolve_path(input_filename: str, reference: str) -> str:
+    """The path `input_filename` is an Excel sheet, and `reference` is a cell
+    value that *might* be a reference to another file. Resolve the path to
+    `reference` and return.
+    """
+    assert str(input_filename).endswith(("xlsx", "csv"))
+
+    # It's not a reference to another file
+    if not reference.endswith(("xlsx", "csv")):
+        return reference
+
+    reference_path = Path(reference)
+
+    # If the reference is e.g. 'C:/Users/USER/files/doe1.xlsx' => full path
+    if reference_path.is_absolute():
+        return str(reference_path.resolve())
+
+    # If the reference is 'doe1.xlsx', then assume the file is in same dir
+    return str(Path(input_filename).parent / reference_path)
+
+
 def _excel2dict_onebyone(
-    input_filename: str, sheetnames: Mapping[str, Any] | None = None
+    input_filename: str | Path,
+    *,
+    gen_input_sheet: str,
+    design_input_sheet: str,
+    default_val_sheet: str,
 ) -> OrderedDict[str, Any]:
     """Reads specification for onebyone design
 
     Args:
-        input_filename(path): path to excel workbook
+        input_filename(str or path): path to excel workbook
+        gen_input_sheet (str): name of general input sheet
+        design_input_sheet (str): name of design input sheet
+        default_val_sheet (str): name of defaul value sheet
         sheetnames (dict): Dictionary of worksheet names to load
             information from. Supported keys: general_input, defaultvalues,
             and designinput.
@@ -210,26 +185,10 @@ def _excel2dict_onebyone(
     Returns:
         OrderedDict on format for DesignMatrix.generate
     """
-
+    input_filename = str(input_filename)
     seedname = "RMS_SEED"
     inputdict: OrderedDict[str, Any] = OrderedDict()
 
-    if not sheetnames or "general_input" not in sheetnames:
-        gen_input_sheet = _find_geninput_sheetname(input_filename)
-    else:
-        gen_input_sheet = sheetnames["general_input"]
-
-    if not sheetnames or "designinput" not in sheetnames:
-        design_inp_sheet = _find_onebyone_input_sheet(input_filename)
-    else:
-        design_inp_sheet = sheetnames["designinput"]
-
-    if not sheetnames or "defaultvalues" not in sheetnames:
-        default_val_sheet = _find_onebyone_defaults_sheet(input_filename)
-    else:
-        default_val_sheet = sheetnames["defaultvalues"]
-
-    # Read general input
     generalinput = pd.read_excel(
         input_filename, gen_input_sheet, header=None, index_col=0, engine="openpyxl"
     )
@@ -238,49 +197,45 @@ def _excel2dict_onebyone(
 
     inputdict["designtype"] = generalinput[1]["designtype"]
 
-    if "rms_seeds" in generalinput[1]:
-        if str(generalinput[1]["rms_seeds"]) == "None":
+    if "rms_seeds" in generalinput.index:
+        rms_seeds = str(generalinput.loc["rms_seeds"].iloc[0])
+        if rms_seeds == "None":
             inputdict["seeds"] = None
         else:
-            inputdict["seeds"] = generalinput[1]["rms_seeds"]
+            inputdict["seeds"] = resolve_path(input_filename, rms_seeds)
     else:
         inputdict["seeds"] = None
 
-    if "repeats" not in generalinput[1]:
+    if "repeats" not in generalinput.index:
         raise LookupError('"repeats" must be specified in general_input sheet')
 
-    inputdict["repeats"] = generalinput[1]["repeats"]
+    inputdict["repeats"] = generalinput.loc["repeats"].iloc[0]
 
-    if "distribution_seed" in generalinput[1]:
-        if str(generalinput[1]["distribution_seed"]) == "None":
+    if "distribution_seed" in generalinput.index:
+        distribution_seed = str(generalinput.loc["distribution_seed"].iloc[0])
+        if distribution_seed == "None":
             inputdict["distribution_seed"] = None
         else:
-            inputdict["distribution_seed"] = generalinput[1]["distribution_seed"]
+            inputdict["distribution_seed"] = int(distribution_seed)
     else:
         inputdict["distribution_seed"] = None
 
-    # Read background
     if "background" in generalinput.index:
+        background = str(generalinput.loc["background"].iloc[0])
         inputdict["background"] = OrderedDict()
-        if generalinput[1]["background"].endswith("csv") or generalinput[1][
-            "background"
-        ].endswith("xlsx"):
-            inputdict["background"]["extern"] = generalinput[1]["background"]
-        elif str(generalinput[1]["background"]) == "None":
+        if background.endswith(("csv", "xlsx")):
+            inputdict["background"]["extern"] = resolve_path(input_filename, background)
+        elif background == "None":
             inputdict["background"] = None
         else:
-            inputdict["background"] = _read_background(
-                input_filename, generalinput[1]["background"]
-            )
+            inputdict["background"] = _read_background(input_filename, background)
     else:
         inputdict["background"] = None
 
-    # Read default values
     inputdict["defaultvalues"] = _read_defaultvalues(input_filename, default_val_sheet)
 
-    # Read input for sensitivities
     inputdict["sensitivities"] = OrderedDict()
-    designinput = pd.read_excel(input_filename, design_inp_sheet, engine="openpyxl")
+    designinput = pd.read_excel(input_filename, design_input_sheet, engine="openpyxl")
     designinput.dropna(axis=0, how="all", inplace=True)
     designinput = designinput.loc[
         :, ~designinput.columns.astype(str).str.contains("^Unnamed")
@@ -297,63 +252,68 @@ def _excel2dict_onebyone(
 
     designinput["sensname"] = designinput["sensname"].ffill()
 
-    # Read dependencies
     if "dependencies" in designinput:
+        valid_deps = designinput[designinput["dependencies"].notna()]
         inputdict["dependencies"] = OrderedDict()
-        for row in designinput.itertuples():
-            if _has_value(row.dependencies):
-                inputdict["dependencies"][row.param_name] = _read_dependencies(
-                    input_filename,
-                    row.dependencies,  # type: ignore[arg-type]
-                    row.param_name,  # type: ignore[arg-type]
-                )
+        for row in valid_deps.itertuples():
+            inputdict["dependencies"][row.param_name] = _read_dependencies(
+                input_filename,
+                str(row.dependencies),
+                str(row.param_name),
+            )
 
-    # Read decimals
     if "decimals" in designinput:
-        inputdict["decimals"] = OrderedDict()
-        for row in designinput.itertuples():
-            if _has_value(row.decimals) and _is_int(row.decimals):  # type: ignore[arg-type]
-                inputdict["decimals"][row.param_name] = int(row.decimals)  # type: ignore[arg-type]
+        # Convert to numeric, then filter for integers
+        numeric_decimals = pd.to_numeric(designinput["decimals"], errors="coerce")
+        mask = numeric_decimals.notna() & (numeric_decimals % 1 == 0)
+
+        valid_decimals = designinput[mask]
+        inputdict["decimals"] = OrderedDict(
+            {
+                row.param_name: int(cast(float, row.decimals))
+                for row in valid_decimals.itertuples()
+            }
+        )
 
     grouped = designinput.groupby("sensname", sort=False)
 
     # Read each sensitivity
     for sensname, group in grouped:
         _check_for_mixed_sensitivities(
-            sensname,  # type: ignore[arg-type]
+            str(sensname),
             group,
         )
 
         sensdict: OrderedDict[str, Any] = OrderedDict()
 
-        if group["type"].iloc[0] == "ref":
-            sensdict["senstype"] = "ref"
+        sens_type = group["type"].iloc[0]
+        if sens_type in {"ref", "background"}:
+            sensdict["senstype"] = sens_type
 
-        elif group["type"].iloc[0] == "background":
-            sensdict["senstype"] = "background"
-
-        elif group["type"].iloc[0] == "seed":
+        elif sens_type == "seed":
             sensdict["seedname"] = seedname
-            sensdict["senstype"] = "seed"
+            sensdict["senstype"] = sens_type
             if _has_value(group["param_name"].iloc[0]):
                 sensdict["parameters"] = _read_constants(group)
             else:
                 sensdict["parameters"] = None
 
-        elif group["type"].iloc[0] == "scenario":
+        elif sens_type == "scenario":
             sensdict = _read_scenario_sensitivity(group)
-            sensdict["senstype"] = "scenario"
+            sensdict["senstype"] = sens_type
 
-        elif group["type"].iloc[0] == "dist":
-            sensdict["senstype"] = "dist"
+        elif sens_type == "dist":
+            sensdict["senstype"] = sens_type
             sensdict["parameters"] = _read_dist_sensitivity(group)
             sensdict["correlations"] = None
             if "corr_sheet" in group:
                 sensdict["correlations"] = _read_correlations(group, input_filename)
 
-        elif group["type"].iloc[0] == "extern":
-            sensdict["extern_file"] = str(group["extern_file"].iloc[0])
-            sensdict["senstype"] = "extern"
+        elif sens_type == "extern":
+            sensdict["extern_file"] = resolve_path(
+                input_filename, str(group["extern_file"].iloc[0])
+            )
+            sensdict["senstype"] = sens_type
             sensdict["parameters"] = list(group["param_name"])
 
         else:
@@ -382,34 +342,31 @@ def _read_defaultvalues(filename: str, sheetname: str) -> OrderedDict[str, Any]:
     Returns:
         OrderedDict with defaultvalues (parameter, value)
     """
-    default_dict: OrderedDict[str, Any] = OrderedDict()
     default_df = pd.read_excel(
         filename, sheetname, header=0, index_col=0, engine="openpyxl"
     )
+
     default_df.dropna(axis=0, how="all", inplace=True)
     default_df = default_df.loc[
         :, ~default_df.columns.astype(str).str.contains("^Unnamed")
     ]
 
-    # Strip spaces before and after parameter names, if they are there
-    # it is probably invisible user errors in Excel.
+    if default_df.empty:
+        return OrderedDict()
 
-    default_df.index = pd.Index(
-        [
-            paramname.strip() if isinstance(paramname, str) else paramname
-            for paramname in default_df.index
-        ]
-    )
-    for row in default_df.itertuples():
-        if str(row[0]) in default_dict:
-            print(
-                f"WARNING: The default value '{row[0]}' "
-                f"is listed twice in the sheet '{sheetname}'. "
-                "Only the first entry will be used in output file"
-            )
-        else:
-            default_dict[str(row[0])] = row[1]
-    return default_dict
+    # Strip leading/trailing spaces from parameter names such that
+    # for example "  PARAM" and "PARAM" are treated as duplicates.
+    default_df.index = default_df.index.str.strip()
+
+    # Check for duplicates and raise error if found
+    duplicates = default_df.index.duplicated(keep=False)
+    if duplicates.any():
+        duplicate_names = default_df.index[duplicates].unique()
+        raise ValueError(
+            f"Duplicate parameter names found in sheet '{sheetname}': "
+            f"{', '.join(duplicate_names)}. All parameter names must be unique."
+        )
+    return OrderedDict(default_df.iloc[:, 0].to_dict())
 
 
 def _read_dependencies(
@@ -720,3 +677,10 @@ def _is_int(teststring: str) -> bool:
         return False  # It was a "number", but it was NaN.
     except ValueError:
         return False
+
+
+def _raise_if_duplicates(container: Sequence[Hashable]) -> None:
+    """Raises a descriptive error if there are duplicates in the container."""
+    duplicates = {k: v for (k, v) in Counter(container).items() if v > 1}
+    if duplicates:
+        raise ValueError(f"Duplicates with counts: {duplicates}")
