@@ -16,7 +16,7 @@ from typing import Any, TypeAlias, cast
 # stdout/stderr during import.
 # https://github.com/cvxpy/cvxpy/issues/2470
 with (
-    open(os.devnull, "w") as devnull,
+    open(os.devnull, "w", encoding="utf-8") as devnull,
     contextlib.redirect_stdout(devnull),
     contextlib.redirect_stderr(devnull),
 ):
@@ -25,12 +25,13 @@ with (
 import numpy as np
 import numpy.typing as npt
 import pandas as pd
-import scipy
 from scipy.stats import qmc
 
 import semeio
 from semeio.fmudesign import design_distributions as design_dist
+from semeio.fmudesign._excel2dict import _raise_if_duplicates
 from semeio.fmudesign.iman_conover import ImanConover
+from semeio.fmudesign.quality_report import QualityReporter
 
 
 def is_consistent_correlation_matrix(matrix: npt.NDArray[Any]) -> bool:
@@ -77,30 +78,38 @@ def nearest_correlation_matrix(
     Parameters
     ----------
     matrix : np.ndarray
-        The matrix that we want to find the nearest correlation matrix to.
-        A square 2-dimensional NumPy ndarray.
+        The matrix that we want to find the nearest positive definite
+        correlation matrix to. A square 2-dimensional NumPy ndarray.
     weights : np.ndarray or None, optional
         An elementwise weighting matrix. A square 2-dimensional NumPy ndarray
         that must have the same shape as `matrix`. The default is None.
     eps : float, optional
-        Tolerance for the optimization solver. The default is 1e-6.
+        Tolerance for the optimization solver and minimum eigenvalue threshold.
+        The result will have all eigenvalues > eps. The default is 1e-6.
     verbose : bool, optional
         Whether to print information from the solver. The default is False.
 
     Returns
     -------
     np.ndarray
-        The correlation matrix that is nearest to the input matrix.
+        The positive definite correlation matrix that is nearest to the input
+        matrix. All eigenvalues will be strictly positive.
 
     Notes
     -----
     This function implements equation (3) in the paper "An Augmented Lagrangian
     Dual Approach for the H-Weighted Nearest Correlation Matrix Problem" by
-    Houduo Qi and Defeng Sun.
+    Houduo Qi and Defeng Sun, with an additional constraint to ensure the
+    result is positive definite (not just positive semidefinite).
         http://www.personal.soton.ac.uk/hdqi/REPORTS/Cor_matrix_H.pdf
     Another useful link is:
         https://nhigham.com/2020/04/14/what-is-a-correlation-matrix/
 
+    The algorithm finds a matrix that is:
+    1. A valid correlation matrix (diagonal elements = 1)
+    2. Positive definite (all eigenvalues > eps)
+    3. Closest to the input matrix in weighted Frobenius norm
+
     Examples
     --------
     >>> X = np.array([[1, 1, 0],
@@ -132,15 +141,16 @@ def nearest_correlation_matrix(
     if not (H.shape == G.shape):
         raise ValueError("Argument `weights` must have same shape as `matrix`.")
 
-    # To constrain Y to be Positive Symmetric Definite (PSD), you need to
+    # To constrain Y to be Positive Semidefinite (PSD), you need to
     # either set PSD=True here, or add the special constraint 'Y >> 0'. See:
     # https://www.cvxpy.org/tutorial/constraints/index.html#semidefinite-matrices
     X = cp.Variable(shape=G.shape, PSD=True)
 
     # Objective and constraints for minimizing the weighted frobenius norm.
-    # This is equation (3) in the paper. We set (X - eps * I) >> 0 as an extra
-    # constraint. Mathematically this is not needed, but numerically it helps
-    # by nudging the solution slightly more, so the minimum eigenvalue is > 0.
+    # This is equation (3) in the paper. We add (X - eps * I) >> 0 as an extra
+    # constraint to ensure the result is positive definite (all eigenvalues > eps).
+    # This constraint makes X - eps*I positive semidefinite, which means
+    # all eigenvalues of X are > eps, ensuring positive definiteness.
     objective = cp.norm(cp.multiply(H, X - G), "fro")
     eps_identity = (eps / G.shape[0]) * 10
     constraints = [cp.diag(X) == 1.0, (X - eps_identity * np.eye(G.shape[0])) >> 0]
@@ -182,18 +192,23 @@ class DesignMatrix:
             or they are read from a file.
     """
 
-    def __init__(self) -> None:
+    def __init__(self, verbosity: int = 0, output_dir: Path | None = None) -> None:
         """
         Placeholders for:
         designvalues: dataframe with parameters that varies
         defaultvalues: dictionary of default/base case values
         backgroundvalues: dataframe with background parameters
         seedvalues: list of seed values
+        verbosity: how much information to print
+        output_dir: where to write debugging output and QC plots
+
         """
         self.designvalues: pd.DataFrame = pd.DataFrame(columns=["REAL"])
         self.defaultvalues: OrderedDict[Hashable, Any] = OrderedDict()
         self.backgroundvalues: pd.DataFrame | None = None
         self.seedvalues: list[int] | None = None
+        self.verbosity: int = verbosity
+        self.output_dir: Path | None = output_dir
 
     def reset(self) -> None:
         """Resets DesignMatrix to empty. Necessary iin case method generate
@@ -214,15 +229,12 @@ class DesignMatrix:
 
         if inputdict["designtype"] != "onebyone":
             raise ValueError(
-                "Generation of DesignMatrix only "
-                "implemented for type onebyone. "
-                "In general_input designtype is "
-                "set to {}".format(inputdict["designtype"])
+                f"Generation of DesignMatrix only implemented for type 'onebyone', not {inputdict['designtype']}"
             )
 
         self.reset()  # Emptying if regenerating matrix
 
-        rng = np.random.RandomState(seed=inputdict.get("distribution_seed"))
+        rng = np.random.default_rng(seed=inputdict.get("distribution_seed"))
 
         self.defaultvalues = inputdict["defaultvalues"]
 
@@ -244,17 +256,20 @@ class DesignMatrix:
         for key in inputdict["sensitivities"]:
             sens = inputdict["sensitivities"][key]
             numreal = sens["numreal"] if "numreal" in sens else inputdict["repeats"]
+
+            print(f"Generating sensitivity : {key}")
+
             if sens["senstype"] == "ref":
                 sensitivity = SingleRealisationReference(key)
                 sensitivity.generate(
-                    range(current_real_index, current_real_index + numreal)
+                    realnums=range(current_real_index, current_real_index + numreal)
                 )
                 current_real_index += numreal
                 self._add_sensitivity(sensitivity)
             elif sens["senstype"] == "background":
                 sensitivity = BackgroundSensitivity(key)
                 sensitivity.generate(
-                    range(current_real_index, current_real_index + numreal)
+                    realnums=range(current_real_index, current_real_index + numreal)
                 )
                 current_real_index += numreal
                 self._add_sensitivity(sensitivity)
@@ -265,10 +280,10 @@ class DesignMatrix:
                     )
                 sensitivity = SeedSensitivity(key)
                 sensitivity.generate(
-                    range(current_real_index, current_real_index + numreal),
-                    sens["seedname"],
-                    self.seedvalues,
-                    sens["parameters"],
+                    realnums=range(current_real_index, current_real_index + numreal),
+                    seedname=sens["seedname"],
+                    seedvalues=self.seedvalues,
+                    parameters=sens["parameters"],
                 )
                 current_real_index += numreal
                 self._add_sensitivity(sensitivity)
@@ -278,9 +293,11 @@ class DesignMatrix:
                     case = sens["cases"][casekey]
                     temp_case = ScenarioSensitivityCase(casekey)
                     temp_case.generate(
-                        range(current_real_index, current_real_index + numreal),
-                        case,
-                        self.seedvalues,
+                        realnums=range(
+                            current_real_index, current_real_index + numreal
+                        ),
+                        parameters=case,
+                        seedvalues=self.seedvalues,
                     )
                     sensitivity.add_case(temp_case)
                     current_real_index += numreal
@@ -288,25 +305,64 @@ class DesignMatrix:
             elif sens["senstype"] == "dist":
                 sensitivity = MonteCarloSensitivity(key)
                 sensitivity.generate(
-                    range(current_real_index, current_real_index + numreal),
-                    sens["parameters"],
-                    self.seedvalues,
-                    sens["correlations"],
+                    realnums=range(current_real_index, current_real_index + numreal),
+                    parameters=sens["parameters"],
+                    seedvalues=self.seedvalues,
+                    corrdict=sens["correlations"],
                     rng=rng,
                 )
                 current_real_index += numreal
                 self._add_sensitivity(sensitivity)
+
             elif sens["senstype"] == "extern":
                 sensitivity = ExternSensitivity(key)
                 sensitivity.generate(
-                    range(current_real_index, current_real_index + numreal),
-                    sens["extern_file"],
-                    sens["parameters"],
-                    self.seedvalues,
+                    realnums=range(current_real_index, current_real_index + numreal),
+                    filename=sens["extern_file"],
+                    parameters=sens["parameters"],
+                    seedvalues=self.seedvalues,
                 )
                 current_real_index += numreal
                 self._add_sensitivity(sensitivity)
             print("Added sensitivity :", sensitivity.sensname)
+
+            # MonteCarloSensitivity is special - it can produce debugging outputs
+            is_montecarlo = isinstance(sensitivity, MonteCarloSensitivity)
+            if is_montecarlo and self.verbosity > 0:
+                sensitivity = cast(MonteCarloSensitivity, sensitivity)
+
+                # Convert parameteters to a plottable string, e.g.
+                # {"COSTS": "Normal(loc=0, scale=1)", ...}
+                qr_vars = {
+                    pname: f"{plist[0]}~("
+                    + ", ".join(
+                        f"dist_param{i}={v}" for (i, v) in enumerate(plist[1], 1)
+                    )
+                    + ")"
+                    for (pname, plist) in sens["parameters"].items()
+                }
+                quality_reporter = QualityReporter(
+                    df=sensitivity.sensvalues, variables=qr_vars
+                )
+
+                # Print to terminal
+                quality_reporter.print_numeric()
+                quality_reporter.print_discrete()
+                for corr_name, df_corr in sensitivity.correlation_dfs_.items():
+                    quality_reporter.print_correlation(corr_name, df_corr)
+
+            if is_montecarlo and self.verbosity > 0 and self.output_dir is not None:
+                sensitivity = cast(MonteCarloSensitivity, sensitivity)
+                output_dir = self.output_dir / key
+                quality_reporter.plot_columns(output_dir=output_dir)
+
+                # Correlations
+                for corr_name, df_corr in sensitivity.correlation_dfs_.items():
+                    quality_reporter.plot_correlation(
+                        corr_name, df_corr, output_dir=output_dir
+                    )
+
+        # Once all sensitivities have been added, complete the work
         if "background" in inputdict:
             self._fill_with_background_values()
         self._fill_with_defaultvalues()
@@ -339,25 +395,22 @@ class DesignMatrix:
             defaultsheet (str): name of excel sheet containing default
                 values (optional, defaults to 'DefaultValues')
         """
-        # pylint: disable=abstract-class-instantiated
-        basename = Path(filename).stem
-        extension = Path(filename).suffix
-        if extension != ".xlsx":
-            filename = basename + ".xlsx"
-            print(
-                "Warning: Output filename did not have extension .xlsx "
-                "but the export format is Excel .xlsx .  "
-                f"Changing outputname to {filename}"
-            )
+        # Create folder for output file
+        Path(filename).parent.mkdir(exist_ok=True, parents=True)
+
+        if Path(filename).suffix != ".xlsx":
+            filename = Path(filename).stem + ".xlsx"
+            print(f"Warning: Missing .xlsx suffix. Changed to: {filename}")
 
         xlsxwriter = pd.ExcelWriter(filename, engine="openpyxl")
         self.designvalues.to_excel(
             xlsxwriter, sheet_name=designsheet, index=False, header=True
         )
         # Default values from OrderdDictionay to pandas dataframe
-        defaults = pd.DataFrame(columns=["defaultparameters", "defaultvalue"])
-        defaults["defaultparameters"] = self.defaultvalues.keys()
-        defaults["defaultvalue"] = self.defaultvalues.values()
+        defaults = pd.DataFrame(
+            data=list(self.defaultvalues.items()),
+            columns=["defaultparameters", "defaultvalue"],
+        )
         defaults.to_excel(
             xlsxwriter, sheet_name=defaultsheet, index=False, header=False
         )
@@ -392,7 +445,7 @@ class DesignMatrix:
                 - str that represents a path to an existing file
             max_reals: Maximum number of seed values to generate or load
         """
-        if seeds in (None, "None"):
+        if seeds in {None, "None"}:
             self.seedvalues = None
             print("seeds is set to None in general_input")
         elif seeds and seeds.lower() == "default":
@@ -411,7 +464,7 @@ class DesignMatrix:
         self,
         back_dict: Mapping[str, Any] | None,
         max_values: int,
-        rng: np.random.RandomState,
+        rng: np.random.Generator,
     ) -> None:
         """Adding background as specified in dictionary.
         Either from external file or from distributions in background
@@ -420,7 +473,7 @@ class DesignMatrix:
         Args:
             back_dict (OrderedDict): how to generate background values
             max_values (int): number of background values to generate
-            rng (numpy.random.RandomState): Random number generator instance
+            rng (numpy.random.Generator): Random number generator instance
         """
         if back_dict is None:
             self.backgroundvalues = None
@@ -441,7 +494,6 @@ class DesignMatrix:
         if self.backgroundvalues is None:
             raise ValueError("No background values available to write to Excel")
 
-        # pylint: disable=abstract-class-instantiated
         xlsxwriter = pd.ExcelWriter(filename, engine="openpyxl")
         self.backgroundvalues.to_excel(
             xlsxwriter, sheet_name=backgroundsheet, index=False, header=True
@@ -458,8 +510,9 @@ class DesignMatrix:
         Args:
             sensitivity of class Scenario, MonteCarlo or Extern
         """
-        existing_values = self.designvalues.copy()
-        self.designvalues = pd.concat([existing_values, sensitivity.sensvalues])
+        existing_values = self.designvalues
+        new_values = sensitivity.sensvalues
+        self.designvalues = pd.concat([existing_values, new_values])
 
     def _fill_with_background_values(self) -> None:
         """Substituting NaNs with background values if existing.
@@ -502,7 +555,7 @@ class DesignMatrix:
                 self.designvalues[key] = self.designvalues[key].fillna(
                     self.defaultvalues[key]
                 )
-            elif key not in ["REAL", "SENSNAME", "SENSCASE", "RMS_SEED"]:
+            elif key not in {"REAL", "SENSNAME", "SENSCASE", "RMS_SEED"}:
                 raise LookupError(f"No defaultvalues given for parameter {key} ")
 
     def _fill_derived_params(self, depend_dict: Mapping[str, Any]) -> None:
@@ -528,7 +581,7 @@ class DesignMatrix:
                         )
 
     def _add_dist_background(
-        self, back_dict: Mapping[str, Any], numreal: int, rng: np.random.RandomState
+        self, back_dict: Mapping[str, Any], numreal: int, rng: np.random.Generator
     ) -> None:
         """Drawing background values from distributions
         specified in dictionary
@@ -536,7 +589,7 @@ class DesignMatrix:
         Args:
             back_dict (OrderedDict): parameters and distributions
             numreal (int): Number of samples to generate
-            rng (numpy.random.RandomState): Random number generator instance
+            rng (numpy.random.Generator): Random number generator instance
         """
         assert isinstance(numreal, int), (
             f"numreal must be an integer, got {type(numreal)} with value {numreal}"
@@ -587,7 +640,10 @@ class DesignMatrix:
 
 class SeedSensitivity:
     """
-    A seed sensitivity is normally the reference for one by one sensitivities
+    A seed sensitivity is normally the reference for one by one sensitivities,
+    which all other sensitivities are compared to. All parameters will be at
+    their default values. Only the RMS_SEED will be varying.
+
     It contains a list of seeds to be repeated for each sensitivity
     The parameter name is hardcoded to RMS_SEED
     It will be assigned the sensname 'p10_p90' which will be written to
@@ -599,7 +655,6 @@ class SeedSensitivity:
 
     """
 
-    # pylint: disable=too-few-public-methods
     def __init__(self, sensname: str) -> None:
         """Initiate method.
 
@@ -607,7 +662,7 @@ class SeedSensitivity:
             sensname (str): Name of sensitivity. Defines SENSNAME in design matrix.
         """
         self.sensname: str = sensname
-        self.sensvalues: pd.DataFrame | None = None
+        self.sensvalues: pd.DataFrame
 
     def generate(
         self,
@@ -664,7 +719,6 @@ class SingleRealisationReference:
 
     """
 
-    # pylint: disable=too-few-public-methods
     def __init__(self, sensname: str) -> None:
         """Initiate.
 
@@ -672,7 +726,7 @@ class SingleRealisationReference:
             sensname (str): Name of sensitivity. Defines SENSNAME in design matrix.
         """
         self.sensname: str = sensname
-        self.sensvalues: pd.DataFrame | None = None
+        self.sensvalues: pd.DataFrame
 
     def generate(self, realnums: range) -> None:
         """Generates realisation number only
@@ -701,7 +755,6 @@ class BackgroundSensitivity:
 
     """
 
-    # pylint: disable=too-few-public-methods
     def __init__(self, sensname: str) -> None:
         """Initiate
 
@@ -709,7 +762,7 @@ class BackgroundSensitivity:
             sensname (str): Name of sensitivity. Defines SENSNAME in design matrix.
         """
         self.sensname: str = sensname
-        self.sensvalues: pd.DataFrame | None = None
+        self.sensvalues: pd.DataFrame
 
     def generate(self, realnums: range) -> None:
         """Generates realisation number only
@@ -744,7 +797,6 @@ class ScenarioSensitivity:
            1-2 cases
     """
 
-    # pylint: disable=too-few-public-methods
     def __init__(self, sensname: str) -> None:
         """
         Args:
@@ -754,7 +806,7 @@ class ScenarioSensitivity:
         self.sensname: str = sensname
         self.case1: ScenarioSensitivityCase | None = None
         self.case2: ScenarioSensitivityCase | None = None
-        self.sensvalues: pd.DataFrame | None = None
+        self.sensvalues: pd.DataFrame
 
     def add_case(self, senscase: "ScenarioSensitivityCase") -> None:
         """
@@ -807,7 +859,6 @@ class ScenarioSensitivityCase:
 
     """
 
-    # pylint: disable=too-few-public-methods
     def __init__(self, casename: str) -> None:
         self.casename: str = casename
         self.casevalues: pd.DataFrame | None = None
@@ -852,10 +903,9 @@ class MonteCarloSensitivity:
             with realisation numbers as index.
     """
 
-    # pylint: disable=too-few-public-methods
     def __init__(self, sensname: str) -> None:
         self.sensname: str = sensname
-        self.sensvalues: pd.DataFrame | None = None
+        self.sensvalues: pd.DataFrame
 
     def _draw_uncorrelated_values(
         self,
@@ -863,12 +913,18 @@ class MonteCarloSensitivity:
         dist_name: str,
         dist_params: Sequence[str],
         numreals: int,
-        rng: np.random.RandomState,
+        rng: np.random.Generator,
     ) -> npt.NDArray[Any] | list[str] | str:
+        # Draw samples in [0, 1)
+        quantiles = design_dist.generate_stratified_samples(numreals=numreals, rng=rng)
+
         try:
             return design_dist.draw_values(
-                dist_name.lower(), dist_params, numreals, rng
+                distname=dist_name.lower(),
+                dist_parameters=dist_params,
+                quantiles=quantiles,
             )
+
         except ValueError as error:
             raise ValueError(
                 f"Problem in sensitivity with sensname {self.sensname} "
@@ -881,7 +937,7 @@ class MonteCarloSensitivity:
         parameters: dict[str, Any],
         seedvalues: Sequence[int] | None,
         corrdict: Mapping[str, Any] | None,
-        rng: np.random.RandomState,
+        rng: np.random.Generator,
     ) -> None:
         """Generates parameter values by drawing from defined distributions.
 
@@ -895,8 +951,12 @@ class MonteCarloSensitivity:
                 If None, parameters are treated as uncorrelated.
         """
         self.sensvalues = pd.DataFrame(columns=list(parameters.keys()), index=realnums)
+        self.correlation_dfs_ = {}  # Store correlation matrices (dataframes)
         numreals = len(realnums)
 
+        if numreals < 0:
+            raise ValueError(f"Got < 0 samples ({numreals=})")
+
         if corrdict is None:
             for key in parameters:
                 self.sensvalues[key] = self._draw_uncorrelated_values(
@@ -911,84 +971,77 @@ class MonteCarloSensitivity:
             df_params["corr_sheet"] = df_params["corr_sheet"].fillna("nocorr")
             df_params.reset_index(inplace=True)
             df_params.rename(columns={"index": "param_name"}, inplace=True)
-            corr_groups = df_params.groupby("corr_sheet")
+            corr_groups = dict(iter(df_params.groupby("corr_sheet")))
+            nocorr = corr_groups.pop("nocorr", pd.DataFrame())
 
-            for corr_group_name, corr_group in corr_groups:
+            for corr_group_name, corr_group in corr_groups.items():
                 corr_group_name = cast(str, corr_group_name)
-                if corr_group_name != "nocorr":
-                    if len(corr_group) == 1:
-                        _printwarning(corr_group_name)
-                    df_correlations = design_dist.read_correlations(
-                        corrdict["inputfile"], corr_group_name
+                if len(corr_group) == 1:
+                    _printwarning(corr_group_name)
+                df_correlations = design_dist.read_correlations(
+                    excel_filename=corrdict["inputfile"], corr_sheet=corr_group_name
+                )
+                multivariate_parameters = list(df_correlations.index.values)
+                correlations = df_correlations.values
+
+                print(
+                    f"Sampling parameters in {corr_group_name!r}: {multivariate_parameters}"
+                )
+
+                if not is_consistent_correlation_matrix(correlations):
+                    print(
+                        f"\nWarning: Correlation matrix {corr_group_name!r} is not consistent"
                     )
-                    multivariate_parameters = df_correlations.index.values
-                    correlations = df_correlations.values
-
-                    # Make correlation matrix symmetric by copying lower triangular part
-                    correlations = np.triu(correlations.T, k=1) + np.tril(correlations)
-
-                    if not is_consistent_correlation_matrix(correlations):
-                        print("\nWarning: Correlation matrix is not consistent")
-                        print("Requirements:")
-                        print("  - Ones on the diagonal")
-                        print("  - Positive semi-definite matrix")
-                        print("\nInput correlation matrix:")
-                        with np.printoptions(
-                            precision=2,
-                            suppress=True,
-                            formatter={"float": lambda x: f"{x:.2f}"},
-                        ):
-                            print(correlations)
-                        correlations = nearest_correlation_matrix(
-                            correlations, weights=None, eps=1e-6, verbose=False
-                        )
-                        print("\nAdjusted to nearest consistent correlation matrix:")
-                        with np.printoptions(
-                            precision=2,
-                            suppress=True,
-                            formatter={"float": lambda x: f"{x:.2f}"},
-                        ):
-                            print(correlations)
-                        print()
-
-                    sampler = qmc.LatinHypercube(
-                        d=len(multivariate_parameters), seed=rng
+                    print("Requirements:")
+                    print("  - Ones on the diagonal")
+                    print("  - Positive semi-definite matrix")
+                    print("\nInput correlation matrix:")
+                    _print_corrmat(df_correlations)
+                    correlations = nearest_correlation_matrix(
+                        correlations, weights=None, eps=1e-6, verbose=False
                     )
-                    lhs_samples = sampler.random(n=numreals)
+                    df_correlations.values[:] = correlations
+                    print("\nAdjusted to nearest consistent correlation matrix:")
+                    _print_corrmat(df_correlations)
 
-                    iman_conover = ImanConover(correlation_matrix=correlations)
-                    correlated_samples = iman_conover(lhs_samples)
+                sampler = qmc.LatinHypercube(d=len(multivariate_parameters), rng=rng)
+                lhs_samples = sampler.random(n=numreals)
 
-                    # Transform uniform correlated samples to normal scores
-                    normalscoresamples = scipy.stats.norm.ppf(correlated_samples)
+                iman_conover = ImanConover(correlation_matrix=correlations)
+                self.correlation_dfs_[corr_group_name] = df_correlations
+                correlated_samples: npt.NDArray[Any] = iman_conover(lhs_samples)
 
-                    for idx, row in corr_group.reset_index(drop=True).iterrows():
-                        if row["param_name"] in multivariate_parameters:
-                            self.sensvalues[row["param_name"]] = (
-                                design_dist.draw_values(
-                                    row["dist_name"].lower(),
-                                    row["dist_params"],
-                                    numreals,
-                                    rng,
-                                    normalscoresamples[:, idx],
-                                )
-                            )
-                        else:
+                for idx, row in corr_group.reset_index(drop=True).iterrows():
+                    idx = cast(int, idx)
+
+                    if row["param_name"] in multivariate_parameters:
+                        if row["param_name"].lower().startswith("const"):
                             raise ValueError(
-                                f"Parameter {row['param_name']} specified with correlation "
-                                f"matrix {corr_group_name} but is not listed in that sheet"
-                            )
-                else:  # group nocorr where correlation matrix is not defined
-                    for _, row in corr_group.iterrows():
-                        self.sensvalues[row["param_name"]] = (
-                            self._draw_uncorrelated_values(
-                                row["param_name"],
-                                row["dist_name"],
-                                row["dist_params"],
-                                numreals,
-                                rng,
+                                "Parameter with const distribution was defined in correlation "
+                                "matrix but const distribution cannot be used with correlation."
                             )
+
+                        self.sensvalues[row["param_name"]] = design_dist.draw_values(
+                            distname=row["dist_name"].lower(),
+                            dist_parameters=row["dist_params"],
+                            quantiles=correlated_samples[:, idx],
                         )
+                    else:
+                        raise ValueError(
+                            f"Parameter {row['param_name']} specified with correlation "
+                            f"matrix {corr_group_name} but is not listed in that sheet"
+                        )
+
+            # Sample every variable without correlation
+            for _, row in nocorr.iterrows():
+                print(f"Sampling parameter {row['param_name']!r}")
+                self.sensvalues[row["param_name"]] = self._draw_uncorrelated_values(
+                    param_name=row["param_name"],
+                    dist_name=row["dist_name"],
+                    dist_params=row["dist_params"],
+                    numreals=numreals,
+                    rng=rng,
+                )
 
         if self.sensname != "background":
             self.sensvalues["REAL"] = realnums
@@ -997,6 +1050,11 @@ class MonteCarloSensitivity:
             if "RMS_SEED" not in self.sensvalues and seedvalues:
                 self.sensvalues["RMS_SEED"] = seedvalues[: len(realnums)]
 
+        null_columns = self.sensvalues.isnull().any(axis=0)
+        if null_columns.any():
+            cols_w_null = list(null_columns.loc[lambda ser: ser].index)
+            raise ValueError(f"Found NaN values in columns: {cols_w_null}")
+
 
 class ExternSensitivity:
     """
@@ -1011,10 +1069,9 @@ class ExternSensitivity:
 
     """
 
-    # pylint: disable=too-few-public-methods
     def __init__(self, sensname: str) -> None:
         self.sensname: str = sensname
-        self.sensvalues: pd.DataFrame | None = None
+        self.sensvalues: pd.DataFrame
 
     def generate(
         self,
@@ -1032,6 +1089,7 @@ class ExternSensitivity:
             parameters (list): list with parameter names
             seeds (str): default or None
         """
+        _raise_if_duplicates(parameters)
         self.sensvalues = pd.DataFrame(columns=parameters, index=realnums)
         extern_values = _parameters_from_extern(filename)
         if len(realnums) > len(extern_values):
@@ -1149,3 +1207,37 @@ def _printwarning(corr_group_name: str) -> None:
         "\n"
         "####################################################\n"
     )
+
+
+def _print_corrmat(df_corrmat: pd.DataFrame) -> None:
+    """Print a correlation matrix.
+
+    Example:
+    >>> values = np.array([[  1, -0,  0.9],
+    ...                    [ -0,  1,    0],
+    ...                    [0.9,  0,    1]])
+    >>> vars_ = ['OWC1', 'OWC2', 'OWC3']
+    >>> df_corrmat = pd.DataFrame(values, index=vars_, columns=vars_)
+    >>> _print_corrmat(df_corrmat)
+    |      |   OWC1 | OWC2   | OWC3   |
+    |:-----|-------:|:-------|:-------|
+    | OWC1 |   1.00 |        |        |
+    | OWC2 |   0.00 | 1.00   |        |
+    | OWC3 |   0.90 | 0.00   | 1.00   |
+    """
+    df_corrmat = df_corrmat.copy()
+    assert np.allclose(df_corrmat.values, df_corrmat.values.T)
+    # Make slightly negative values positive
+    values = df_corrmat.to_numpy()
+    mask = np.isclose(values, 0)
+    values[mask] = np.abs(values[mask])
+    df_corrmat.values[:] = values
+
+    # Remove upper triangular part for prettier printing
+    formatter = lambda x: np.format_float_positional(
+        x, precision=2, unique=True, min_digits=2
+    )
+    mask = np.triu(np.ones_like(df_corrmat, dtype=bool), k=1)
+    df_display = df_corrmat.astype(float).map(formatter)
+    df_display[mask] = ""
+    print(df_display.to_markdown(floatfmt=".2f"))
